from typing import Dict, List

import yaml


class Yml:
    @staticmethod
    def load_docs(path: str) -> List[Dict[any, any]]:
        docs = []
        with open(path) as f:
            data = yaml.load_all(f, Loader=yaml.FullLoader)
            for doc in data:
                if doc is None:
                    # Empty block
                    continue
                docs.append(doc)
        return docs



class MissingParam(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class MissingVar(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class ConfigError(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class SkipObject(Exception):
    """
    Indicates that the object in the current context should not be deployed
    """
    def __init__(self, msg: str):
        super().__init__(msg)



def remove_prefix(text, prefix):
    """
    Removeprefix for python <3.9
    :param text: Text
    :param prefix: Prefix to be removed
    :return: Text
    """
    if text.startswith(prefix):
        return text[len(prefix):]
    return text



import base64
import os
from typing import Dict

import Crypto.Hash.SHA256
from Crypto import Random
from Crypto.Cipher import AES
from Crypto.Hash import SHA512
from Crypto.Protocol.KDF import PBKDF2

from octoploy.k8s.SecretObj import SecretObj
from octoploy.utils.Yml import Yml
from octoploy.utils.YmlWriter import YmlWriter


class AESCipher(object):

    def __init__(self, key):
        self.bs = AES.block_size
        salt = b'octoployPepper!!'
        dv = PBKDF2(key, salt, 64, count=100000, hmac_hash_module=SHA512)
        self.key = dv[:32]

    def encrypt(self, raw: str) -> str:
        raw_bytes = raw.encode('utf-8')
        # Append 32 bytes of message hash to detect successful decryption
        validation = Crypto.Hash.SHA256.SHA256Hash().new(raw_bytes).digest()
        raw_bytes += validation

        raw_bytes = self._pad(raw_bytes)
        iv = Random.new().read(self.bs)
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return base64.b64encode(iv + cipher.encrypt(bytes(raw_bytes))).decode('utf-8')

    def decrypt(self, enc: str) -> str:
        enc = base64.b64decode(enc.encode('utf-8'))

        iv = enc[:self.bs]
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        raw_bytes = self._unpad(cipher.decrypt(enc[self.bs:]))
        payload = raw_bytes[:-32]

        payload_hash = Crypto.Hash.SHA256.SHA256Hash().new(payload).digest()
        expected_hash = raw_bytes[-32:]
        if payload_hash != expected_hash:
            raise ValueError('Could not decrypt value')

        return payload.decode('utf-8')

    def _pad(self, s: bytes) -> bytearray:
        """
        Pads the argument to be a multiple of the block size
        :param s: Bytes
        :return: Padded bytes
        """
        s = bytearray(s)
        padding_count = (self.bs - len(s) % self.bs)
        for x in range(padding_count):
            s.append(padding_count)

        return s

    @staticmethod
    def _unpad(s: bytes) -> bytes:
        """
        Removes the padding
        :param s: Padded bytes
        :return: Raw bytes
        """
        return s[:-ord(s[len(s) - 1:])]


class Encryption:
    KEY_ENV = 'OCTOPLOY_KEY'
    CRYPT_PREFIX = 'OctoCrypt!'

    def __init__(self):
        self._cipher = None

    def encrypt(self, raw: str) -> str:
        cipher = self._get_cipher()
        return cipher.encrypt(raw)

    def decrypt(self, enc: str) -> str:
        cipher = self._get_cipher()
        return cipher.decrypt(enc)

    def _get_cipher(self):
        if self._cipher is None:
            key = os.environ.get(self.KEY_ENV)
            if key is None:
                raise ValueError(f'Environment {self.KEY_ENV} is not defined. The key is required for '
                                 f'de/encryption')
            self._cipher = AESCipher(key)
        return self._cipher


class YmlEncrypter:
    """
    Encrypts k8s secret objects
    """

    def __init__(self, path: str):
        self.path = path
        self._encryption = Encryption()

    def encrypt(self):
        docs = Yml.load_docs(self.path)
        did_find_secrets = False
        for doc in docs:
            try:
                secret = SecretObj(doc)
                did_find_secrets = True
            except ValueError:
                # Not a secret
                continue
            self._encrypt(secret.base64_data)
            self._encrypt(secret.string_data)

        if not did_find_secrets:
            raise ValueError(f'Did not find a single secret in {self.path}')
        with open(self.path, 'w') as file:
            YmlWriter.dump_all(docs, file)

    def _encrypt(self, data: Dict[str, any]):
        for key, value in data.items():
            if not isinstance(value, str) or value.startswith(Encryption.CRYPT_PREFIX):
                continue
            data[key] = Encryption.CRYPT_PREFIX + self._encryption.encrypt(value)



from unittest import TestCase

from octoploy.utils.Errors import MissingVar

from octoploy.config.AppConfig import AppConfig


class AppConfigTest(TestCase):
    def test_missing_name(self):
        config = AppConfig('', None)
        config.data = {
            'forEach': [
                {}
            ]
        }
        try:
            config.get_for_each()
            self.fail('No exception')
        except MissingVar:
            pass



import io
import time
from typing import Optional, List
from unittest.mock import Mock

from octoploy.api.Oc import Oc

OCTOPLOY_KEY = 'key123'


class DummyCmd:
    args = []
    namespace: None
    stdin: None

    def __init__(self, args, stdin, namespace):
        self.args = args
        self.stdin = stdin
        self.namespace = namespace

    def __repr__(self):
        out = f'{self.args}'
        if self.stdin is not None:
            out += f' <- {self.stdin}'
        return out


class DummyK8sApi(Oc):
    def __init__(self):
        super().__init__()
        self.commands = []
        self.responds = []
        self._respond_not_found = False

    def not_found_by_default(self):
        self._respond_not_found = True

    def respond(self, args: List[str], stdout: '', error=None):
        self.responds.append((args, stdout, error))

    def _exec(self, args, print_out: bool = False, stdin: str = None, namespace: Optional[str] = None) -> str:
        self.commands.append(DummyCmd(args, stdin, namespace))
        for response in self.responds:
            expected_args = response[0]
            stdout = response[1]
            error = response[2]
            if expected_args == args:
                if error is not None:
                    raise error
                return stdout
        if self._respond_not_found and args[0] == 'get':
            raise Exception('NotFound')
        return '{}'


class TestHelper:
    @staticmethod
    def mock_popen_success(mock_popen, stdout=b'output', stderr=b'error', sleep_time: int = 0,
                           return_code: int = 0):
        """
        Modifies the Popen "poll" call to always return 0

        :param mock_popen: mock of Popen
        :param stdout: Std out that the process should return
        :param stderr: Std err that the process should return
        :param sleep_time: Number of seconds how long the poll() call should wait
        :param return_code: Return code of the process
        """

        stdout_stream = io.BytesIO(stdout)
        stderr_stream = io.BytesIO(stderr)

        def poll_sleep():
            if sleep_time > 0:
                time.sleep(sleep_time)
            return return_code

        process_mock = Mock()
        attrs = {'communicate.return_value': (stdout, stderr),
                 'stdout': stdout_stream,
                 'stderr': stderr_stream,
                 'poll': poll_sleep}
        process_mock.configure_mock(**attrs)
        # always return success for process execution
        mock_popen.return_value = process_mock



from typing import Dict, Optional


class BaseObj:
    api_version: str
    kind: str
    name: Optional[str]
    namespace: Optional[str]
    metadata: Dict[str, any]

    def __init__(self, data: Dict[str, any]):
        self.data = data
        self.kind = data['kind']
        self.api_version = data['apiVersion']

        self.metadata = data.get('metadata', {})
        self.name = self.metadata.get('name', None)
        self.namespace = self.metadata.get('namespace', None)

    def get_fqn(self) -> str:
        """
        Returns the complete name of this object, including the group and kind
        :return: Name
        """
        group = self.get_group()
        if group is None:
            return f'{self.kind}/{self.name}'

        return f'{self.kind}.{group}/{self.name}'

    def get_group(self) -> Optional[str]:
        """
        Returns the group part of the api version.
        For example "networking.k8s.io"
        :return: Group or None
        """
        segments = self.api_version.split('/')
        if len(segments) < 2:
            # Plain version
            return None
        return segments[0]

    def is_kind(self, kind: str) -> bool:
        if self.kind is None:
            return False
        return self.kind.lower() == kind.lower()

    def require_kind(self, kind: str):
        if not self.is_kind(kind):
            raise ValueError(f'Object is not of kind {kind}')



from __future__ import annotations

from typing import Dict, List, Optional

import yaml

from octoploy.api.Oc import K8sApi
from octoploy.k8s.BaseObj import BaseObj
from octoploy.utils.Log import Log
from octoploy.utils.YmlWriter import YmlWriter


class ObjectState:
    def __init__(self):
        self.context = ''
        self.name = ''
        self.kind = ''
        self.hash = ''
        self.namespace: Optional[str] = None
        self.visited = False
        """
        Transient flag to indicate if the object has been visited by octoploy
        """

    def update_from_key(self, key: str):
        segments = key.split('/')
        count = len(segments)
        if count > 0:
            self.context = segments[0]
        if count > 1:
            self.namespace = segments[1]
        if count > 2:
            self.kind = segments[2]
        if count > 3:
            self.name = segments[3]

    def parse(self, data: Dict[str, str]) -> ObjectState:
        self.context = data['context']
        self.namespace = data['namespace']
        self.kind = data['kind']
        self.name = data['name']
        self.hash = data.get('hash', '')
        return self

    def to_dict(self) -> Dict[str, str]:
        return {
            'name': self.name,
            'hash': self.hash,
            'kind': self.kind,
            'context': self.context,
            'namespace': self.namespace,
        }

    def get_key(self) -> str:
        return f'{self.context}/{self.namespace}/{self.kind}/{self.name}'


class StateTracking(Log):
    """
    Stores the objects that have been deployed with octoploy.
    This allows octoploy to detect renamed / deleted objects.
    """
    CM_NAME = 'octoploy-state'
    _k8s_api: K8sApi
    _state: Dict[str, ObjectState]

    def __init__(self, api: K8sApi, name_suffix: str = ''):
        super().__init__()
        self._k8s_api = api
        self._cm_name = self.CM_NAME + name_suffix
        self._state = {}

    def restore(self, namespace: str):
        item = self._k8s_api.get(f'ConfigMap/{self._cm_name}', namespace=namespace)
        if item is None:
            return
        cm = item.data
        state_data_str = cm.get('data', {}).get('state', '')
        state_data = yaml.safe_load(state_data_str)
        if state_data is None:
            return

        for state_obj in state_data:
            object_state = ObjectState().parse(state_obj)
            self._state[object_state.get_key()] = object_state

    def store(self, namespace: str):
        self.log.debug(f'Persisting state in ConfigMap {self._cm_name}')

        states = []
        for object_state in self._state.values():
            states.append(object_state.to_dict())

        data = {
            'kind': 'ConfigMap',
            'apiVersion': 'v1',
            'metadata': {
                'name': self._cm_name
            },
            'data': {
                'state': YmlWriter.dump(states)
            }
        }
        yml = YmlWriter.dump(data)
        self._k8s_api.apply(yml, namespace=namespace)

    def move(self, source: str, dest: str):
        if source.count('/') != dest.count('/'):
            raise ValueError('Source and destination point to different path depths')

        for value in list(self._state.values()):
            key = value.get_key()
            if not key.startswith(source):
                continue
            target = key.replace(source, dest)
            self.log.info(f'Moving {key} to {target}')
            value.update_from_key(target)

            if key in self._state:
                del self._state[key]
            self._state[target] = value

    def get_not_visited(self, context: str) -> List[ObjectState]:
        """
        Returns all objects which have not been visited
        :param context: Context
        :return: Objects
        """
        items = []
        for object_state in self._state.values():
            if not object_state.visited and object_state.context == context:
                items.append(object_state)
        return items

    def get_state(self, context_name: str, k8s_object: BaseObj) -> Optional[ObjectState]:
        state = self._k8s_to_state(context_name, k8s_object)
        return self._state.get(state.get_key())

    def visit(self, context_name: str, k8s_object: BaseObj, hash_val: str):
        """
        Marks the given object as "visited".
        If the object is not yet in the state it will be added
        :param context_name: Name of the state context
        :param k8s_object: Kubernetes object
        :param hash_val: The new hash value of the object.
        """
        state = self._k8s_to_state(context_name, k8s_object)
        existing_state = self._state.get(state.get_key())
        if existing_state is None:
            state.hash = hash_val
            self._state[state.get_key()] = state
            return
        existing_state.hash = hash_val
        existing_state.visited = True

    def remove(self, object_state: ObjectState):
        del self._state[object_state.get_key()]

    @staticmethod
    def _k8s_to_state(context_name: str, k8s_object: BaseObj) -> ObjectState:
        api_version = k8s_object.api_version
        kind = k8s_object.kind
        name = k8s_object.name

        state = ObjectState()
        state.namespace = k8s_object.namespace
        state.context = context_name
        state.api_version = api_version
        state.kind = kind
        state.name = name
        state.visited = True
        return state



from __future__ import annotations

import re
from typing import Optional, Dict, List, Set
from typing import TYPE_CHECKING

from octoploy.processing.TreeWalker import TreeWalker, TreeProcessor
from octoploy.utils.Log import Log

if TYPE_CHECKING:
    from octoploy.config.BaseConfig import BaseConfig
from octoploy.utils.Errors import MissingParam


class YmlTemplateProcessor(Log, TreeProcessor):
    """
    Processes yml files by replacing any string placeholders.
    """

    VAR_PATTERN = re.compile(r'\${(.+?)}')
    KEY_FIELD_MERGE: str = '_merge'

    _replacements: Dict[str, any] = {}

    def __init__(self, config: BaseConfig):
        super().__init__()
        self._missing_vars = []  # type: List[str]
        """
        List of all variables which have not been replace because
        there was no value defined for them
        """
        self._config = config  # type: BaseConfig
        self._parent = None  # type: Optional[YmlTemplateProcessor]
        self._child = None  # type: Optional[YmlTemplateProcessor]

        self._replacements = {}

    def parent(self, template_processor: YmlTemplateProcessor):
        """
        Inherits all replacements from the given processor.
        If the same value is defined in this and the parent, the definition of this processor will override
        the parent definition

        :param template_processor: Child processor
        """
        if self._parent is not None:
            raise ValueError('Parent processor already defined')
        self._parent = template_processor

    def child(self, template_processor: YmlTemplateProcessor):
        """
        Inherits all replacements from the given processor.
        If the same value is defined in this and the child, the definition of the child processor will override
        this definition

        :param template_processor: Child processor
        """
        if self._child is not None:
            raise ValueError('Child processor already defined')
        self._child = template_processor

    def process(self, data: dict):
        """
        Processes the app data

        :param data: Data of the app, the data will be modified in place
        :raise MissingParam: Gets raised if at least one parameter is not defined
        """
        self._load_replacements()

        # Now replace any placeholders in the actual data tree
        walker = TreeWalker(self)
        walker.walk(data)

        # self._walk_dict(replacements, data)

        # Check if any of the missing vars are declared as "params"
        # (aka are required)
        if len(self._missing_vars) > 0:
            missing_params = []
            params = self._get_params()
            for missing in self._missing_vars:
                if missing not in params:
                    continue
                missing_params.append(missing)
            if len(missing_params) > 0:
                raise MissingParam('The following params are not defined: ' + str(missing_params))
            self.log.warning('The following vars are not defined: ' + str(self._missing_vars))

    def _get_params(self) -> Set[str]:
        """
        Returns all defined params
        :return: Param names
        """
        params = set()
        if self._parent is not None:
            params.update(self._parent._get_params())
        params.update(self._config.get_params())
        if self._child is not None:
            params.update(self._child._get_params())
        return params

    def process_object(self, data: Dict[str, any], parent: Dict[str, any], key: str) -> Optional[Dict[str, any]]:
        if key == self.KEY_FIELD_MERGE:
            # Remove the item from the parent
            # and merge it into the parent itself
            del parent[key]
            TreeWalker(self).walk(data)
            parent.update(data)
            return None
        return data

    def process_str(self, value: str, parent: Dict[str, any], key: str) -> any:
        if key == self.KEY_FIELD_MERGE:
            # Remove the item from the parent
            # and merge it into the parent itself
            del parent[key]
            new_value = self._replace(value)
            if isinstance(new_value, str):
                raise ValueError(f'No replacement found for {new_value}')
            parent.update(new_value)
            return None

        return self._replace(value)

    def _replace(self, item: str) -> any:
        """
        Replaces the given variable with the actual value
        :param item: Variable in the format "${KEY}"
        :return: Value or item if no replacement was found
        """
        for variable, value in self._replacements.items():
            if item == '${' + variable + '}':
                # Item only contains a tag, simple replace (non textual)
                return value
            # The variable tag is surrounded by other str or other tags
            item = item.replace('${' + variable + '}', str(value))

        # Search for any missing variables
        if isinstance(item, str):
            self._missing_vars.extend(self.VAR_PATTERN.findall(item))
        return item

    def _load_replacements(self):
        """
        Loads all available replacements
        """
        self._replacements = self._get_replacements()
        self._resolve_refs(self._replacements)

    def _get_replacements(self) -> Dict[str, any]:
        """
        Returns all replacements handled by this processor, including all parent variables
        :return: Replacements
        """
        replacements = {}
        if self._parent is not None:
            replacements.update(self._parent._get_replacements())
        replacements.update(self._config.get_replacements())
        if self._child is not None:
            replacements.update(self._child._get_replacements())
        return replacements

    def _resolve_refs(self, data: Dict[str, any]):
        """
        Resolves any references in the given replacements (in place)
        :param data: Replacements
        """

        found_var = True
        while found_var:
            found_var = False
            for key, value in data.items():
                if isinstance(value, dict):
                    # A replacement might be an object
                    # So we need to walk through every item and see if there is something to replace
                    self._resolve_refs(value)
                    continue
                if not isinstance(value, str):
                    continue

                # The replacement value might refer to another variable
                for variable_name in self.VAR_PATTERN.findall(value):
                    new_value = data.get(variable_name, self._replacements.get(variable_name))
                    if new_value is None:
                        self.log.warning('Missing referenced variable: ' + variable_name)
                        continue
                    if value == '${' + variable_name + '}':
                        # Replace the entire value since the replacement value only consists of the ${} tag
                        data[key] = new_value
                        found_var = True
                        continue

                    data[key] = value.replace('${' + variable_name + '}', str(new_value))
                    found_var = True



import hashlib

from octoploy.api.Oc import K8sApi
from octoploy.config.Config import RootConfig, AppConfig, RunMode
from octoploy.k8s.BaseObj import BaseObj
from octoploy.state.StateTracking import StateTracking
from octoploy.utils.Log import Log
from octoploy.utils.YmlWriter import YmlWriter


class K8sObjectDeployer(Log):
    """
    Deploys k9s object
    """

    HASH_ANNOTATION = 'yml-hash'
    """
    Name of the annotation that stores the hash
    """
    _root_config: RootConfig
    _app_config: AppConfig
    _api: K8sApi
    _state: StateTracking

    def __init__(self, root_config: RootConfig, k8sapi: K8sApi, app_config: AppConfig, mode: RunMode = RunMode()):
        super().__init__()
        self._root_config = root_config
        self._app_config = app_config
        self._api = k8sapi
        self._mode = mode
        self._state = root_config.get_state()

    def select_context(self):
        """
        Selects the cluster context
        """
        context = self._root_config.get_kubectl_context()
        if context is not None:
            self._api.switch_context(context)

    def deploy_object(self, data: dict):
        """
        Deploy the given object (if a deployment required, otherwise does nothing)
        :param data: Data which should be deployed
        """
        # Sort the content so it's always reproducible
        str_repr = YmlWriter.dump(data)
        hash_val = hashlib.md5(str_repr.encode('utf-8')).hexdigest()

        k8s_object = BaseObj(data)
        # An object might be in a different namespace than the current context
        namespace = k8s_object.namespace
        if namespace is None:
            namespace = self._root_config.get_namespace_name()
        k8s_object.namespace = namespace  # Make sure the object points to the correct namespace

        item_path = k8s_object.get_fqn()
        current_object = self._api.get(item_path, namespace=namespace)
        if current_object is None:
            self.log.info(f'{item_path} will be created')

        current_hash = None
        if current_object is not None:
            obj_state = self._state.get_state(self._app_config.get_name(), k8s_object)
            if obj_state is not None and obj_state.hash != '':
                current_hash = obj_state.hash
            else:  # Fallback to old hash location
                current_hash = current_object.get_annotation(self.HASH_ANNOTATION)

        if current_object is not None and current_hash is None:
            # Item has not been deployed with octoploy, but it does already exist
            self.log.warning(f'{item_path} has no state annotation, assuming no change required')
            self._state.visit(self._app_config.get_name(), k8s_object, hash_val)
            return

        if current_hash == hash_val:
            self._state.visit(self._app_config.get_name(), k8s_object, hash_val)
            self.log.debug(f"{item_path} hasn't changed")
            return

        if current_object is not None:
            self.log.info(f'{item_path} will be updated')

        if self._mode.plan:
            self._state.visit(self._app_config.get_name(), k8s_object, hash_val)
            return

        self._api.apply(str_repr, namespace=namespace)
        self._state.visit(self._app_config.get_name(), k8s_object, hash_val)

        if k8s_object.is_kind('ConfigMap'):
            self._reload_config()

    def delete_abandoned_objects(self):
        """
        Deletes all objects which are not anymore included in the
        current app config
        """
        abandoned = self._state.get_not_visited(self._app_config.get_name())
        for item in abandoned:
            namespace = item.namespace
            if namespace is None:
                namespace = self._root_config.get_namespace_name()
            item_path = item.kind + '/' + item.name
            self.log.info(f'{item_path} will be deleted')
            if self._mode.plan:
                continue

            self._api.delete(item_path, namespace=namespace)
            self._state.remove(item)

    def _reload_config(self):
        """
        Tries to reload the configuration for the app
        """
        reload_actions = self._app_config.get_reload_actions()
        for action in reload_actions:
            action.run(self._api)



import setuptools
import codecs
import os

with open('README.md', 'r') as fh:
    long_description = fh.read()


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setuptools.setup(
    name='octoploy',
    version=get_version('octoploy/__init__.py'),
    author='davidgiga1993',
    author_email='david@dev-core.org',
    description='Simple kubernetes / openshift templating engine with templating, libraries, state tracking',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/davidgiga1993/octoploy',
    packages=setuptools.find_packages(),
    python_requires='>=3.8',
    install_requires=['pyyaml', 'pycryptodome'],
    entry_points={
        'console_scripts': ['octoploy=octoploy.octoploy:main'],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries'
    ],
)



from unittest import TestCase

from octoploy.utils.YmlWriter import YmlWriter


class YmlWriterTest(TestCase):
    def test_long_str(self):
        data = {
            'a': 'this is a very long string which should not be split up into strange multiple lines with backslash'
        }
        out_str = YmlWriter.dump(data)
        self.assertEqual('"a": "this is a very long string which should not be split up into strange multiple lines with backslash"\n', out_str)


from __future__ import annotations

import os
from typing import Optional, Dict, List

from octoploy.api.Oc import Oc, K8, K8sApi
from octoploy.config.AppConfig import AppConfig
from octoploy.config.BaseConfig import BaseConfig
from octoploy.processing.DataPreProcessor import DataPreProcessor, OcToK8PreProcessor
from octoploy.processing.YmlTemplateProcessor import YmlTemplateProcessor
from octoploy.state.StateTracking import StateTracking
from octoploy.utils.Errors import ConfigError


class RunMode:
    def __init__(self):
        self.out_file = None  # type: Optional[str]
        """
        Yml output file
        """

        self.dry_run = False
        """
        True if no OC should be called
        """

        self.plan = False
        """
        True if changes should be previewed
        """


class RootConfig(BaseConfig):
    """
    Root configuration for a project (aka a collection of apps inside a single context/namespace)
    """

    def __init__(self, config_root: str, path: str):
        super().__init__(path)
        self._config_root = config_root
        self._oc = None
        self._library = None  # type: Optional[RootConfig]

        inherit = self.data.get('inherit')
        if inherit is not None:
            # Use a library
            parent_dir = os.path.abspath(os.path.join(path, os.pardir, os.pardir))
            lib_dir = os.path.join(parent_dir, inherit)
            if not os.path.isdir(lib_dir):
                raise FileNotFoundError('Library not found: ' + lib_dir)
            self._library = RootConfig.load(lib_dir)
            if not self._library.is_library():
                raise ConfigError('Project ' + inherit + ' referenced as library but is not a library')

        state_name = self.data.get('stateName', '')
        self._state = StateTracking(self.create_api(), state_name)

    @classmethod
    def load(cls, path: str) -> RootConfig:
        return RootConfig(path, os.path.join(path, '_root.yml'))

    def get_state(self) -> StateTracking:
        return self._state

    def initialize_state(self, run_mode: RunMode):
        if run_mode.dry_run:
            return
        self._state.restore(self.get_namespace_name())

    def persist_state(self, run_mode: RunMode):
        if run_mode.dry_run or run_mode.plan:
            return
        self._state.store(self.get_namespace_name())

    def get_config_root(self) -> str:
        return self._config_root

    def is_library(self) -> bool:
        """
        Indicates if this collection is a library
        """
        return self.data.get('type', '') == 'library'

    def get_pre_processor(self) -> DataPreProcessor:
        """
        Returns the pre processor for the current config
        """
        mode = self._get_mode()
        if mode == 'k8s' or mode == 'k8s':
            return OcToK8PreProcessor()
        return DataPreProcessor()

    def _get_mode(self) -> str:
        return self.data.get('mode', 'k8s')

    def create_api(self) -> K8sApi:
        """
        Creates a new openshift / k8s client
        :return: Client
        """
        if self._oc is not None:
            return self._oc

        mode = self._get_mode()
        if mode == 'oc':
            oc = Oc()
        elif mode == 'k8s' or mode == 'k8':
            oc = K8()
        else:
            raise ValueError(f'Invalid mode: {mode}')
        self._oc = oc
        return oc

    def get_namespace_name(self) -> Optional[str]:
        """
        Returns the namespace name of this project

        :return: Name or null for libraries or if the namespace should not be changed
        """
        return self.data.get('namespace', self.data.get('project'))

    def get_kubectl_context(self) -> Optional[str]:
        """
        Returns the configuration context name
        :return: Name or null if the current context should be used
        """
        return self.data.get('context')

    def get_template_processor(self) -> YmlTemplateProcessor:
        root_processor = super().get_template_processor()
        if self._library is not None:
            processor = self._library.get_template_processor()
            root_processor.parent(processor)
        return root_processor

    def get_replacements(self) -> Dict[str, str]:
        """
        Returns all variables which are available for the yml files
        :return: Key, value map
        """
        items = super().get_replacements()
        name = self.get_namespace_name()
        if name is not None:
            items.update({
                'OC_PROJECT': name,
                'NAMESPACE': name
            })
        return items

    def load_app_configs(self) -> List[AppConfig]:
        """
        Loads all app configurations available in this project
        :return:
        """
        items = []
        names = set()
        for dir_item in os.listdir(self._config_root):
            path = os.path.join(self._config_root, dir_item)
            if not os.path.isdir(path):
                continue
            try:
                app_config = self.load_app_config(dir_item)
            except FileNotFoundError:
                # Index file missing
                continue

            if app_config.is_template() or not app_config.enabled():
                # Silently skip
                continue
            app_name = app_config.get_name()
            if app_name in names:
                raise ValueError(f'The app name {app_name} has already been used')
            names.add(app_name)
            items.append(app_config)
        if self._library is not None:
            items.extend(self._library.load_app_configs())
        return items

    def load_app_config(self, name: str) -> AppConfig:
        folder_path = os.path.join(self._config_root, name)
        if not os.path.isdir(folder_path):
            if self._library is not None:
                return self._library.load_app_config(name)
            raise FileNotFoundError('App folder not found: ' + folder_path)

        index_file = os.path.join(folder_path, '_index.yml')
        if not os.path.isfile(index_file):
            # Index file missing
            raise FileNotFoundError('No index yml file found: ' + index_file)

        variables = self.get_replacements()
        return AppConfig(folder_path, index_file, variables)
